from operator import itemgetter
from typing import Literal,List,Annotated
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage
from langchain_core.pydantic_v1 import BaseModel
from typing_extensions import TypedDict
from langchain_core.tools import tool,Tool
from langchain_core.runnables import RunnablePassthrough,RunnableLambda,RunnableParallel
from langgraph.graph import START,END,StateGraph,MessagesState
from langgraph.graph.message import add_messages
from langgraph.checkpoint.sqlite import SqliteSaver

from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph.state import CompiledStateGraph

from ylz_utils.file import FileLib
from ylz_utils.data import StringLib,Color
#from ylz_utils.langchain import LangchainLib
class State(TypedDict):
    messages: Annotated[list,add_messages]
    ask_human: bool

class RequestAssistance(BaseModel):
    """Escalate the conversation to an expert. Use this if you are unable to assist directly or if the user requires support beyond your permissions.

    To use this function, relay the user's 'request' so the expert can provide the right guidance.
    """
    request: str    

class GraphLib():
    def __init__(self,langchainLib,db_conn_string=":memory:"):
        self.langchainLib = langchainLib
        #self.websearch_tool = self.set_websearch_tool(websearch_key)        
        #self.ragsearch_tool = langchainLib.get_ragsearch_tool(retriever)
        self.websearch_tool = None
        self.ragsearch_tool = None
        self.python_repl_tool = langchainLib.get_python_repl_tool()
        self.memory = SqliteSaver.from_conn_string(db_conn_string)

        self.llm_with_tools = None
    def set_dbname(self,dbname):
        "checkpoint.sqlite"
        self.memory = SqliteSaver.from_conn_string(dbname)
    def set_ragsearch_tool(self,retriever):
        name = "rag_searcher"
        description = "一个有用的工具用来从本地知识库中获取信息。你总是利用这个工具优先从本地知识库中搜索有用的信息"
        self.ragsearch_tool = self.langchainLib.get_ragsearch_tool(retriever,name,description)
    def set_websearch_tool(self,websearch_key):
        self.websearch_tool = self.langchainLib.get_websearch_tool(websearch_key)
    
    def translate_to_en(self,state:State):
        return state
        content = state["messages"][-1].content
        prompt = self.langchainLib.get_prompt(
"""
将以下句子翻译成英文,注意不要遗漏任何信息，数字不是整数的话要保留所有整数和小数，不要翻译任何公司或人名
然后一步一步获得最终答案
""",
        use_chat=False)
        #llm = self.langchainLib.get_llm(model = "llama3-groq-70b-8192-tool-use-preview")
        #llm = self.langchainLib.get_llm(model = "llama3-70b-8192")
        llm = self.langchainLib.get_llm(key = "LLM.TOGETHER")
        chain = prompt | llm
        response = chain.invoke({"input":content})
        return {"messages":[response],"ask_human":False}
    
    def chatbot(self,state: State):
        response = self.llm_with_tools.invoke(state["messages"])
        ask_human = False
        if response.tool_calls and response.tool_calls[0]["name"] == RequestAssistance.__name__:
            ask_human = True
        return {"messages":[response],"ask_human":ask_human}
    
    def router(self,state:MessagesState)-> Literal["tools","__end__"]:
            messages = state["messages"]
            tool_calls = messages[-1].tool_calls
            if tool_calls:
                return "tools"
            else:
                return END
    def create_response(self, response: str, ai_message: AIMessage):
        return ToolMessage(
            content = response,
            tool_call_id = ai_message.tools_calls[0]["id"]
        )
    
    def human_node(self, state: State):
        new_messages = []
        if not isinstance(state["messages"][-1],ToolMessage):
            new_messages.append(
                self.create_response("No response from human.",state["messages"][-1])
            )
        return {
            "messages": new_messages,
            "ask_human": False
        }
    
    def select_next_node(self, state:State) -> Literal["human","tools","__end__"]:
        if state["ask_human"]:
            return "human"
        return tools_condition(state)
     
    def get_graph(self,llm_key=None,llm_model=None):
        llm = self.langchainLib.get_llm(llm_key,llm_model)
        tools = [
            self.python_repl_tool,
        ]
        if self.websearch_tool:
            tools.append(self.websearch_tool)
        if self.ragsearch_tool:
            tools.append(self.ragsearch_tool)
        # tools = [
        #        Tool(name="websearch",
        #             description = "the tool when you can not sure,please search from the internet",
        #             func = websearch
        #        ),
        #        Tool(name="python_repl",
        #             description="when you need to calculation, use python repl tool to execute code ,then return the result to AI.",
        #             func=self.python_repl_tool)
        #      ]
        self.llm_with_tools = llm.bind_tools(tools+[ RequestAssistance ])      

        workflow = StateGraph(State)
        workflow.add_node("translate",self.translate_to_en)
        workflow.add_node("chatbot",self.chatbot)
        workflow.add_node("tools",ToolNode(tools=tools))
        workflow.add_node("human", self.human_node)
        
        workflow.add_conditional_edges("chatbot", self.select_next_node,
                                       {"human":"human","tools":"tools","__end__":"__end__"})
        
        workflow.add_edge(START, "translate")
        workflow.add_edge("translate","chatbot")
        workflow.add_edge("tools","chatbot")
        workflow.add_edge("human","chatbot")

        graph = workflow.compile(checkpointer=self.memory,
                                 interrupt_before= ["human"])

        return graph    

    def graph_stream(self,graph,message,thread_id="default-default",system_message=None):    
        messages = []
        if system_message:
            messages.append(SystemMessage(content=system_message))
        if message:
            messages.append(HumanMessage(content=message))
            values = {"messages":messages}          
        else:
            values = None
        events = graph.stream( values,
                                config = {"configurable":{"thread_id":thread_id}},
                                stream_mode = "values")
        for event in events:
            if "messages" in event:
                message = event['messages'][-1]
                if isinstance(message,AIMessage):
                    if message.tool_calls:
                        print("AI:",f'使用{Color.GREEN}{message.tool_calls[0]["name"]}{Color.RESET},调用参数:{Color.GREEN}{message.tool_calls[0]["args"]}{Color.RESET}')
                    else:
                        print("AI:",message.content,
                              f'[model:{Color.LYELLOW}{message.response_metadata["model_name"]}{Color.RESET},token:{Color.LYELLOW}{message.usage_metadata["total_tokens"]}{Color.RESET}]')
                elif isinstance(message,ToolMessage):
                    print("    Tool:",message.content)
    
    def graph_get_state_history(self,graph,thread_id="default-default"):
        state_history = graph.get_state_history(config = {"configurable":{"thread_id":thread_id}} )
        for state in state_history:
            print("Num Messages: ", len(state.values["messages"]), "Next: ", state.next)
            print("-" * 80)
        return state_history
    
    def graph_get_state(self,graph:CompiledStateGraph,thread_id="default-default"):
        state = graph.get_state(config =  {"configurable":{"thread_id":thread_id}})
        return state
    
    def graph_update_state(self,graph:CompiledStateGraph,thread_id,values,as_node = None):
        if not as_node:
            graph.update_state(config = {"configurable":{"thread_id":thread_id}},values=values)

    def export_graph(self,graph:CompiledStateGraph):
        FileLib.writeFile("graph.png",graph.get_graph().draw_mermaid_png(),mode="wb")    



from ylz_utils.langchain.documents.dir_loader import DirLoader
from ylz_utils.langchain.documents.docx import DocxLib
from ylz_utils.langchain.documents.image import ImageLib
from ylz_utils.langchain.documents.pdf import PdfLib
from ylz_utils.langchain.documents.pptx import PptxLib
from ylz_utils.langchain.documents.url import UrlLib


class DocumentLib():
   def __init__(self,langchainLib):
      self.dir = DirLoader(langchainLib)
      self.url = UrlLib(langchainLib)
      self.docx = DocxLib(langchainLib)
      self.pptx = PptxLib(langchainLib)
      self.image = ImageLib(langchainLib)
      self.pdf = PdfLib(langchainLib)


from langchain_community.document_loaders.image import UnstructuredImageLoader

class ImageLib():
     def __init__(self,langchainLib):
          self.langchainLib = langchainLib
     def loader(self,filename):
         "./example_data/layout-parser-paper-screenshot.png"
         return UnstructuredImageLoader(filename)


from typing import Type
from langchain_core.tools import Tool
from langchain_experimental.utilities import PythonREPL
from langchain_core.pydantic_v1 import BaseModel,Field
class PythonREPLArgSchema(BaseModel):
    ''' input str for python repl tool'''
    command: str = Field(description="the command to execute use python repl")

class PythonREPLTool():
    def __init__(self,langchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def get_tool(self,name=None):
        # You can create the tool to pass to an agent
        python_repl = PythonREPL()
        name = name or "python_repl"
        #So You are not skilled at dealing with computation issues,All question about calculate shoud use this tool when you get all argument value.
        repl_tool = Tool(
            name=name,
            description="""
The tool is a Python shell. Use this to execute python commands. Input should be a valid python command. If you want to see the output of a value, you should print it out with `print(...)`.""",
            args_schema = PythonREPLArgSchema ,
            func=python_repl.run
        )
        return repl_tool




import asyncio
from pydoc import writedoc
from playwright.async_api import async_playwright, Browser, Page
import time
import shlex
from .langchain import LangchainLib
from .soup import SoupLib
from tqdm import tqdm
import logging

class PlaywrightLib:
    def __init__(self, browser_type="chromium", headless=False):
        self.browser_type = browser_type
        self.headless = headless
        self.playwright = None
        self.browser = None
        self.page = None
    async def __aenter__(self):
        self.playwright = await async_playwright().start()
        self.__browser_type = getattr(self.playwright, self.browser_type)
        self.browser: Browser = await self.__browser_type.launch(headless=self.headless)
        self.page: Page = await self.browser.new_page()
        return self
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.browser.close()
        await self.playwright.stop()
    async def goto(self, url,timeout=60000,wait_until="load",start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        await self.page.goto(url,timeout=timeout,wait_until=wait_until)
        bool(end_log) and logging.info(end_log)

    async def click(self, selector,start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        await self.page.click(selector)
        bool(end_log) and logging.info(end_log)

    async def fill(self, selector, text,start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        await self.page.fill(selector, text)
        bool(end_log) and logging.info(end_log)

    async def get_html(self, selector=None, start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        if bool(selector):
            elements = await self.page.query_selector_all(f"xpath={selector}")
            content = content = await asyncio.gather(
                *[element.evaluate('el => el.innerHTML') for element in elements]
            )
        else:
            content = [await self.page.content()]
        bool(end_log) and logging.info(end_log)
        return content
    async def replace_html(self, selector=None, innerHTML='<div>HELLO...</div>',start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        if bool(selector):
            elements = await self.page.query_selector_all(f"xpath={selector}")
            await asyncio.gather(
                *[element.evaluate(f'el => el.innerHTML="{innerHTML}"') for element in elements]
            )
            # for element in elements:
            #     print(element)
            #     await self.page.evaluate(
            #         f" (element) => element.innerHTML = ```{innerHTML}``` ", element
            #     )
            # self.page.evaluate(
            #     """(selector) => {
            #         const element = document.querySelector(selector);
            #         if (element) {
            #             element.innerHTML = '<div> hello </div>';
            #         }
            #     }""",
            #     selector,  # 将选择器传递给 JavaScript 函数
            # )
            # element = self.page.locator(selector)
            # print("*"*50,element[0].inner_html())
            # element[0].inner_html = '<div> hello </div>' 
        else:
            await self.page.set_content(innerHTML)
        bool(end_log) and logging.info(end_log)
    async def selector_exists(self, selector) -> bool:
        element = self.page.locator(selector)
        return await element.count() > 0
    async def wait_for_selector(self, selector, timeout=5000,start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        await self.page.wait_for_selector(selector, timeout=timeout)
        bool(end_log) and logging.info(end_log)

    async def wait_for_load_state(self, state,start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        await self.page.wait_for_load_state(state)
        bool(end_log) and logging.info(end_log)
    
    async def text_content(self, selector, start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        text = await self.page.text_content(selector)
        bool(end_log) and logging.info(end_log)
        return text
    async def alltext_content(self, selector, start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        elements = await self.page.query_selector_all(f"xpath={selector}")
        textes = list(map(lambda element:element.text_content(), elements))
        bool(end_log) and logging.info(end_log)
        return textes
    
    async def screenshot(self, path="screenshot.png"):
        await self.page.screenshot(path=path)

    def wait(self, milliseconds,start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        time.sleep(milliseconds / 1000)  # 将毫秒转换为秒
        bool(end_log) and logging.info(end_log)
    
    async def close(self):
        await self.browser.close()
        await self.playwright.stop()

    async def get_attributes(self, selector, attribute_names, start_log="",end_log=""):
        bool(start_log) and logging.info(start_log)
        if type(attribute_names)!=list and type(attribute_names)!=tuple:
            attribute_names = [attribute_names]
        attribute_values = await self.page.evaluate(
            """([xpath, attributes]) => {
                const elements = document.evaluate(xpath, document, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
                const values = [];
                for (let i = 0; i < elements.snapshotLength; i++) {
                    const element = elements.snapshotItem(i);
                    attrMap = {}
                    for (let attribute of attributes) {
                        attrMap[attribute] = element.getAttribute(attribute);
                    }
                    values.push(attrMap);
                }
                return values;
            }""",
            (selector,attribute_names),
            
        )
        bool(end_log) and logging.info(end_log)
        return attribute_values
    async def get_html_between_xpaths(self, xpath1, xpath2,include_start=True,include_end=False):
        """获取两个 XPath 之间的 HTML 片段。

        Args:
            xpath1: 第一个 XPath 表达式。
            xpath2: 第二个 XPath 表达式。
            include_start: 是否包含第一个 XPath 匹配的元素。
            include_end: 是否包含第二个 XPath 匹配的元素。
        Returns:
            一个字符串，表示两个 XPath 之间的 HTML 片段，
            如果找不到任何匹配的元素，则返回 None。
        """

        html = await self.page.evaluate(
            """([xpath1, xpath2,include_start,include_end]) => {
                const startNode = document.evaluate(xpath1, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
                const endNode = document.evaluate(xpath2, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;

                if (!startNode || !endNode) {
                    return null;
                }

                let currentNode = startNode;
                let html = "";
                while (currentNode && currentNode !== endNode) {
                    if (currentNode == startNode && !include_start) {
                       currentNode = currentNode.nextSibling;
                       continue
                    }
                    html += currentNode.outerHTML;
                    currentNode = currentNode.nextSibling;
                }
                if (include_end) {
                    html += endNode.outerHTML; // 包含结束节点
                }

                return html;
            }
            """,
            (xpath1,xpath2,include_start,include_end),
        )        
        return html
    
     
async def main():
    #url = "https://global.alipay.com/docs/ac/ams/payment_agreement"
    #url = "https://global.alipay.com/docs/ac/ams/payment_cashier"
    #url = "https://global.alipay.com/docs/ac/ams/supply_evidence"
    url  = "https://global.alipay.com/docs/ac/ams/api_fund"
    langchainLib = LangchainLib()
    llm = langchainLib.get_chatopenai_llm(
            base_url="https://api.siliconflow.cn/v1",
            api_key="api_key",
            model="alibaba/Qwen1.5-110B-Chat",temperature=0)
    systemPromptText = """你是专业的金融技术领域专家,同时也是互联网信息化专家。熟悉蚂蚁金服的各项业务,擅长这些方面的技术文档的翻译。
    现在请将下面的HTML格式文档全部翻译成中文,输出HTML文档。
    要求:
        1、不要有遗漏,简单明了。
        2、特别不要遗漏嵌套的HTML的语法
        3、禁止翻译代码中的JSON的key
        4、保留所有原始的HTML格式
        5、检查翻译的结果,以确保语句通顺
    \n\n"""
    prompt = langchainLib.get_prompt(textwrap.dedent(systemPromptText))
    chain = prompt | llm
    # html = readFile("/Users/youht/source/python/translate/translate/test/Response header.html")
    # async with PlaywrightWrapper(headless=False) as pw:
    #     soup = pw.html2soup(html)
    #     newsoup,attribute_dict = pw.hash_attribute(soup) 
    #     writeFile("newsoup.html",newsoup.prettify())
    #     html_cn = chain.invoke({"input":newsoup.prettify()})
    #     soup_cn = pw.to_soup(html_cn.content)
    #     dumpJson("attribute_dict.json",attribute_dict)
    #     soup_cn= pw.unhash_attribute(soup_cn,attribute_dict)
    #     writeFile("test_cn.html",soup_cn.prettify())
        
    # html_cn = 
    # writeFile("test.html",html_cn.content)
    
    async with PlaywrightLib(headless=False) as pw:
        await pw.goto(url,start_log="开始加载页面",end_log="页面加载完成",wait_until="domcontentloaded")

        # pw.wait_for_selector("#Requestparameters")
        # pw.click('//div[contains(@class,"sandboxSwitch")]//span[text()="Sample Codes"]',start_log='sample1 code')
        # textes = pw.alltext_content('//div[@id="ace-editor"]//div[@class="ace_content"]//div[contains(@class,"ace_text-layer")]',
        #                          end_log="获取脚本文本")
        # print(textes)
        # pw.click('//div[contains(@class,"sandboxSwitch")]//span[text()="Run in Sandbox"]',start_log='sample2 code')
        # textes = pw.alltext_content('//div[@id="ace-editor"]//div[@class="ace_content"]//div[contains(@class,"ace_text-layer")]',
        #                          end_log="获取脚本文本")
        # print(textes)
        
        # pw.replace_html("//span[text()='Structure']")    
        attr = await pw.get_attributes("//nav//a",["href","title"])
        print(attr)
        pw.wait(3000,start_log="等待3秒",end_log="等待3秒结束")
        #await pw.replace_html("//h3",innerHTML='<div> 临时替换 </div>')
        html_cn = ""
        with tqdm(total= len(attr)) as pbar:
            for index,item in enumerate(attr):
                start = None
                end = None
                title = None
                if index < len(attr) - 1:
                    start = item.get("href",None)
                    end = attr[index + 1].get("href",None)
                if not start or not end:
                    break  
                start = start[1:]
                end = end[1:]
                title = item.get("title")
                html_snip = await pw.get_html_between_xpaths(f"//*[@id='{start}']",f"//*[@id='{end}']",include_start=True,include_end=False)
                soup_snip = SoupLib.html2soup(html_snip)
                attribute_dict = SoupLib.hash_attribute(soup_snip) 
                hash_html_snip_cn = chain.invoke({"input":pw.soup2html(soup_snip)})
                hash_soup_snip_cn = SoupLib.html2soup(hash_html_snip_cn.content)
                SoupLib.unhash_attribute(hash_soup_snip_cn,attribute_dict)
                html_snip_cn = SoupLib.soup2html(hash_soup_snip_cn)
                html_cn += html_snip_cn
                pbar.update(1)
        FileLib.writeFile("test_cn.html",html_cn)
        # pw.click("//div[@id='Requestparameters']//button//span[contains(text(),'Show all')]",start_log="点击Req Show all按钮")
        # pw.click("//div[@id='Responseparameters']//button//span[contains(text(),'Show all')]",start_log="点击Res Show all按钮")
        # pw.wait_for_selector("//div[@id='Requestparameters']//button//span[contains(text(),'Hide all')]",start_log="定位Req Hide all按钮")
        # pw.wait_for_selector("//div[@id='Responseparameters']//button//span[contains(text(),'Hide all')]",start_log="定位Req Hide all按钮")
        # # 等待页面加载完成 (可选，但建议使用)
        # pw.wait_for_load_state("load") 
        #pw.wait(60000,start_log="等待60秒",end_log="等待结束")
        # writeFile("test.html",pw.get_html())
        await pw.close()

    '''
        # 要翻译的部分
        //article[@class="ant-typography"]//section
        # sandboxSwitch span按钮
        //div[contains(@class,"sandboxSwitch")]//span[text()="Sample Codes"]
        //div[contains(@class,"sandboxSwitch")]//span[text()="Run in Sandbox"]
        #脚本文本
        //div[@id="ace-editor"]//div[@class="ace_content"]//div[contains(@class,"ace_text-layer")]
        #定位id
        //*[@id="3RxeL"]
        //*[@id="d8Mc5"]
    '''

if __name__ == "__main__":
   asyncio.run(main()) 




from langchain.tools.retriever import create_retriever_tool

class RagSearchTool():
    def __init__(self,langchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def get_tool(self,retriever,name,description):
        return create_retriever_tool(retriever,name,description)


from ylz_utils.file import FileLib
from ylz_utils.crypto import HashLib
from ylz_utils.config import Config
from ylz_utils.langchain import LangchainLib
from ylz_utils.playwright import PlaywrightLib
from ylz_utils.soup import SoupLib
from ylz_utils.data import StringLib,Color,JsonLib,UrlLib

__all__ = ['FileLib','SoupLib','HashLib','Config','PlaywrightLib','LangchainLib','StringLib','Color','JsonLib','UrlLib']



import logging
from logging.handlers import TimedRotatingFileHandler

from ylz_utils import Config
from ylz_utils import StringLib

def init(args):
    project_name = args.project_name
    # 设置logger
    log_level = logging.INFO if args.log_level=="INFO" else logging.DEBUG
    log_file = args.log_name if args.log_name else f"{project_name}.log"

    logger = logging.getLogger()
    logger.setLevel(log_level)

    file_handler = TimedRotatingFileHandler(
        filename= log_file,
        when="midnight",  # 每天午夜滚动
        interval=1,  # 滚动间隔为 1 天
        backupCount=7,  # 保留 7 天的日志文件
        encoding='utf-8'
    )
    #file_handler = logging.FileHandler("task.log")
    #formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatter = logging.Formatter('%(asctime)s %(name)s [pid:%(process)d] [%(threadName)s] [%(levelname)s] %(message)s')
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # looging
    StringLib.logging_in_box(str(args),"*")

    #设置config
    Config.init(project_name) 


from ylz_utils.config import Config
from ylz_utils.data import StringLib

import time
import random

from langchain_openai import ChatOpenAI
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_community.chat_models import QianfanChatEndpoint
from langchain_google_genai import GoogleGenerativeAIEmbeddings
from langchain_ollama import ChatOllama
from langchain_community.chat_message_histories import SQLChatMessageHistory,ChatMessageHistory
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.runnables import ConfigurableFieldSpec

from gradio_client import Client,file

class LLMLib():
    llms:list = [] 
    default_llm_key = None
    chat_dbname = None
    def __init__(self):
        self.config = Config.get()
        self.regist_llm()
        if self.config.get("LLM.DEFAULT"):
            default_llm_key = self.config.get("LLM.DEFAULT")
            self.set_default_llm_key(default_llm_key)
        else:
            self.clear_default_llm_key()
        #self.add_plugins()
    def set_default_llm_key(self,key):
        llms = [item for item in self.llms if item['type']==key]
        if llms:
            self.default_llm_key = key
        else:
            self.clear_default_llm_key()
    def get_default_llm_key(self):
        return self.default_llm_key
    def clear_default_llm_key(self):
        self.default_llm_key=None
    def set_dbname(self,dbname):
        self.chat_dbname = dbname
    def get_user_session_history(self, user_id: str, conversation_id: str):
        if self.chat_dbname:
            return SQLChatMessageHistory(f"{user_id}--{conversation_id}", f"sqlite:///{self.chat_dbname}")
        raise Exception("you must give the chat_dbname . please use set_dbname first!")
        
    def get_chat(self,llm,prompt,history_messages_key = "history",input_key = "input"):
        return RunnableWithMessageHistory(
            prompt | llm,
            self.get_user_session_history,
            input_messages_key = input_key,
            history_messages_key= history_messages_key,
            history_factory_config=[
                    ConfigurableFieldSpec(
                        id="user_id",
                        annotation=str,
                        name="User ID",
                        description="Unique identifier for the user.",
                        default="",
                        is_shared=True,
                    ),
                    ConfigurableFieldSpec(
                        id="conversation_id",
                        annotation=str,
                        name="Conversation ID",
                        description="Unique identifier for the conversation.",
                        default="",
                        is_shared=True,
                    ),
                ]
        )

    def get_llm(self,key=None, model=None, full=False, delay=10)->ChatOpenAI | ChatOllama:
        if full:
            return self.llms
        if self.llms:
            if not key:
                if self.default_llm_key:
                    llms = [item for item in self.llms if item['type']==self.default_llm_key]
                else:
                    llms = self.llms
            else:
                if key == "LLM.GEMINI":
                    #macos不支持
                    llms=[]
                else:
                    llms = [item for item in self.llms if item['type']==key]
            if llms:
                while True:
                    llm = random.choice(llms) 
                    #print(llm.get('last_used',0),time.time() - llm.get('last_used'),llm['api_key'])   
                    over_delay = (time.time() - llm.get('last_used',0)) > delay # 3秒内不重复使用
                    if not over_delay:
                        continue
                    if not model and llm['llm']==None:
                        llm['model'] = llm['default_model']
                    elif model and model != llm.get('model'):
                        llm['model'] = model
                        llm['llm'] = None 
                    if not llm.get('llm'):
                        llm_type = llm.get('type')
                        if llm_type == 'LLM.GEMINI':
                            try:
                                StringLib.logging_in_box(f"google_api_key={llm.get('api_key')},model={llm.get('model')}")
                                llm['llm'] = ChatGoogleGenerativeAI(
                                    google_api_key=llm.get('api_key'), 
                                    model=llm.get('model'),
                                    temperature= llm.get('temperature'))
                            except:
                                raise Exception(f"请确保{llm_type}_API_KEYS环境变量被正确设置")                                
                        elif llm_type == 'LLM.QIANFAN':
                            try:
                                StringLib.logging_in_box(f"qianfan_api_key={llm.get('api_key')},qianfan_sec_key={llm.get('sec_key')},model={llm.get('model')}")
                                llm['llm'] = QianfanChatEndpoint(
                                    qianfan_ak= llm.get('api_key'),
                                    qianfan_sk= llm.get('sec_key'),
                                    model= llm.get('model'))
                            except:
                                raise Exception(f"请确保{llm_type}_API_KEYS和{llm_type}_SEC_KEYS环境变量被正确设置")
                        elif llm_type == 'LLM.OLLAMA':
                            try:
                                StringLib.logging_in_box(f"ollama_api_key={llm.get('api_key')},model={llm.get('model')}")
                                llm['llm'] = ChatOllama(model= llm.get('model'),
                                                        temperature= llm.get('temperature'),
                                                        keep_alive=llm.get('keep_alive'))
                            except:
                                raise Exception(f"请确保{llm_type}_API_KEYS环境变量被正确设置")
                            
                        else:
                            try:
                                llm['llm'] = ChatOpenAI(
                                    base_url=llm.get('base_url'),
                                    api_key= llm.get('api_key'),
                                    model= llm.get('model'),
                                    temperature= llm.get('temperature')
                                    )
                            except:
                                raise Exception(f"请确保{llm_type}_API_KEYS环境变量被正确设置")                                
                            
                    llm['used'] = llm.get('used',0) + 1 
                    llm['last_used'] = time.time()
                    return llm['llm'] 
        if self.default_llm_key:
            raise Exception(f"请确保{self.default_llm_key}_API_KEYS环境变量被正确设置,然后调用regist_llm注册语言模型")
        else:
            raise Exception(f"请确保<LLM PROVIDER>_API_KEYS环境变量被正确设置,然后调用regist_llm注册语言模型")        
    def regist_llm(self):
        defaults = {"LLM.TOGETHER":
                      {"model":"Qwen/Qwen1.5-72B-Chat","temperature":0},
                   "LLM.SILICONFLOW":
                      {"model":"alibaba/Qwen1.5-110B-Chat","temperature":0},
                   "LLM.GROQ":
                      {"model":"llama-3.1-70b-versatile","temperature":0},
                   "LLM.GEMINI":
                      {"model":"gemini-pro","temperature":0},
                    "LLM.DEEPSEEK":
                      {"model":"deepseek-chat","temperature":1},
                    "LLM.QIANFAN":
                      {"model":"Yi-34B-Chat","temperature":0.7},
                    "LLM.OLLAMA":
                      {"model":"llama3.1","temperature":0.7},
                    "LLM.MOONSHOT":
                      {"model":"moonshot-v1-8k","temperature":0.3}
                  }
        for key in defaults:
            default = defaults[key]
            llm = self.config.get(key)
            if not llm:
                continue
            base_url = llm.get("BASE_URL")
            api_keys = llm.get("API_KEYS")
            keep_alive = llm.get("KEEP_ALIVE")
            if api_keys:
                api_keys = api_keys.split(",")
                if not api_keys[-1]:
                    api_keys.pop()
            else:
                api_keys = []

            sec_keys = llm.get("SEC_KEYS")
            if sec_keys:
                sec_keys = sec_keys.split(",")
                # 防止最后
                if not sec_keys[-1]:
                    sec_keys.pop()
            else:
                sec_keys = []
            model= llm.get("MODEL") if llm.get("MODEL") else default['model']
            temperature = llm.get("TEMPERATURE") if llm.get("TEMPERATURE") else default['temperature']
            for idx, api_key in enumerate(api_keys):
                if key == "LLM.QIANFAN" and len(api_keys) == len(sec_keys):
                    sec_key = sec_keys[idx]
                else:
                    sec_key = ""
                self.llms.append({
                    "llm": None,
                    "type": key,
                    "base_url":base_url,
                    "api_key":api_key,
                    "sec_key": sec_key,
                    "default_model": model,
                    "model":model,
                    "temperature":temperature,
                    "keep_alive":keep_alive,
                    "used":0,
                    "last_used": 0 
                })
    def get_opengpt4o_client(self,api_key):
        client = Client("KingNish/OpenGPT-4o")
        return client
    def opengpt4o_predict(self, client,prompt="hello",imageUrl=None):
        imageFile = file(imageUrl) if imageUrl else None
        result = client.predict(
                image3=imageFile,
                prompt3=prompt,
                api_name="/predict"
        )
        return result
    def opengpt4o_chat(self,client,prompt="hello",imageUrls=[],temperature=0.5,webSearch=True):
        imageFiles = [file(imageUrl) for imageUrl in imageUrls] if imageUrls else []
        result = client.predict(
            message={"text":prompt,"files":imageFiles},
            request="idefics2-8b-chatty",
            param_3="Top P Sampling",
            param_4=temperature,
            param_5=4096,
            param_6=1,
            param_7=0.9,
            param_8=webSearch,
            api_name="/chat"
        )
        return result
    def get_chatopenai_llm(self,base_url,api_key,model,temperature=0.7):
        llm = ChatOpenAI(
                base_url=base_url,
                api_key=api_key,
                model=model,
                temperature=temperature
                )
        return llm



import logging
import inspect
import json
from datetime import datetime

from langchain_openai import ChatOpenAI

from langchain_ollama import ChatOllama
from langchain_core.messages import SystemMessage,HumanMessage

from langchain.memory import ConversationBufferMemory

from langchain.docstore.document import Document

from langchain_community.tools.tavily_search.tool import TavilySearchResults

from langchain_core.prompts import ChatPromptTemplate,PromptTemplate,SystemMessagePromptTemplate
from langchain_core.prompt_values import StringPromptValue,ChatPromptValue

from langchain_core.runnables import RunnableParallel, RunnablePassthrough, RunnableLambda,Runnable
from operator import itemgetter

from langgraph.graph import START,END,MessageGraph,StateGraph
from langgraph.prebuilt import ToolNode

from ylz_utils.langchain.agents import AgentLib
from ylz_utils.langchain.graph import GraphLib
from ylz_utils.langchain.llms import LLMLib
from ylz_utils.langchain.embeddings import EmbeddingLib
from ylz_utils.langchain.documents import DocumentLib
from ylz_utils.langchain.prompts import PromptLib
from ylz_utils.langchain.output_parsers import OutputParserLib
from ylz_utils.langchain.splitters import SplitterLib
from ylz_utils.langchain.tools import ToolLib
from ylz_utils.langchain.vectorstores import VectorstoreLib

from ylz_utils.file import FileLib
from ylz_utils.config import Config
from ylz_utils.data import StringLib,Color

class LangchainLib():
    def __init__(self):
        self.config = Config.get()
        self.memory = ConversationBufferMemory()

        self.llmLib = LLMLib()
        self.get_chat = self.llmLib.get_chat
        self.get_llm = self.llmLib.get_llm

        self.embeddingLib = EmbeddingLib()
        self.get_embedding = self.embeddingLib.get_embedding

        self.promptLib = PromptLib()
        self.get_prompt = self.promptLib.get_prompt

        self.outputParserLib = OutputParserLib()
        self.get_outputParser = self.outputParserLib.get_outputParser
        
        self.vectorstoreLib = VectorstoreLib(self)
        
        self.toolLib = ToolLib(self)
        self.get_websearch_tool = self.toolLib.web_search.get_tool
        self.get_ragsearch_tool = self.toolLib.rag_search.get_tool
        self.get_python_repl_tool = self.toolLib.python_repl.get_tool

        self.splitterLib = SplitterLib()
        self.get_textsplitter = self.splitterLib.get_textsplitter
        self.split_markdown_docs = self.splitterLib.split_markdown_docs

        self.documentLib = DocumentLib(self)
        self.load_url_and_split_markdown = self.documentLib.url.load_and_split_markdown

        self.agentLib = AgentLib(self)
        self.get_agent = self.agentLib.get_agent

        self.graphLib = GraphLib(self)
        self.get_graph = self.graphLib.get_graph
        
    def add_plugins(self,debug=False):
        plugins = [{"class":ChatOpenAI,"func":ChatOpenAI.invoke},
                   {"class":ChatOpenAI,"func":ChatOpenAI.ainvoke},
                   {"class":ChatOpenAI,"func":ChatOpenAI.stream},
                   {"class":ChatOpenAI,"func":ChatOpenAI.astream},
                   {"class":TavilySearchResults,"func":TavilySearchResults.invoke}
        ]
        def get_wrapper(func):
            logging.info(f"增加{func.__qualname__}的插件!!")
            def sync_wrapper(self, *args,**kwargs):
                if isinstance(args[0],StringPromptValue):
                    text = args[0].text
                    args[0].text = f"现在是北京时间:{datetime.now().strftime('%Y-%m-%d %H:%M:%S %A')}\n{text}"
                elif isinstance(args[0],ChatPromptValue):
                    chatPromptValue = args[0]
                    chatPromptValue.messages.insert(0,SystemMessage(f"现在是北京时间:{datetime.now().strftime('%Y-%m-%d %H:%M:%S %A')}"))
                elif isinstance(args[0],str):
                    new_args = list(args).copy()
                    text = new_args[0]
                    new_args[0] = f"现在是北京时间:{datetime.now().strftime('%Y-%m-%d %H:%M:%S %A')}\n{text}"                    
                    #print("?"*10,new_args,"现在仍使用原args")
                if debug:
                    StringLib.logging_in_box("提示词-->"+str(args))
                return func(self, *args, **kwargs)
            async def async_wrapper(self, *args,**kwargs):
                if isinstance(args[0],StringPromptValue):
                    text = args[0].text
                    args[0].text = f"现在是北京时间:{datetime.now().strftime('%Y-%m-%d %H:%M:%S %A')}\n{text}"
                elif isinstance(args[0],ChatPromptValue):
                    chatPromptValue = args[0]
                    chatPromptValue.messages.insert(0,SystemMessage(f"现在是北京时间:{datetime.now().strftime('%Y-%m-%d %H:%M:%S %A')}"))
                if debug:
                    StringLib.logging_in_box("提示词-->"+str(args))
                return await func(self, *args, **kwargs)
            if  inspect.iscoroutinefunction(func):
                return async_wrapper
            else:
                return sync_wrapper
        for item in plugins:
            func = item["func"]
            func_name = func.__name__
            cls = item["class"]
            setattr(cls,func_name,get_wrapper(func))
    



import asyncio
import logging
import argparse

from ylz_utils.cli.init import init

from ylz_utils.cli.reset import reset
from ylz_utils.cli.start import start
from ylz_utils.cli.serve import serve

def run():
    main()

def main():
    parser = argparse.ArgumentParser(description = "测试工具")
    parser.add_argument("--project_name",type=str,default="ylz_utils",help="project名称")
    parser.add_argument("--config_name",type=str,default="config.yaml",help="config名称")
    parser.add_argument("--log_level",type=str,default="INFO",choices=["INFO","DEBUG"],help="日志级别,默认:INFO")
    parser.add_argument("--log_name",type=str,default="ylz_utils.log",help="日志文件名称")
    
    subparsers = parser.add_subparsers(dest="command", required=True, help="可以使用的子命令")
    
    reset_parser = subparsers.add_parser("reset", help="执行初始化")

    start_parser = subparsers.add_parser("start", help="启动测试")
    start_parser.add_argument("--mode",type=str,
                              required=True,
                              choices=["llm","chat","prompt","loader","runnable","tools","rag","outputParser","graph","agent","info"],
                              help="测试内容")
    start_parser.add_argument("--llm_key",type=str,help="语言模型标识，例如：LLM.DEEPSEEK")
    start_parser.add_argument("--embedding_key",type=str,help="嵌入模型标识，例如：EMBEDDING.TOGETHER")
    start_parser.add_argument("--llm_model",type=str,help="语言模型model")
    start_parser.add_argument("--embedding_model",type=str,help="嵌入模型model")
    start_parser.add_argument("--message",type=str,help="input message")
    start_parser.add_argument("--user",type=str,help="user_id,example: alice")    
    start_parser.add_argument("--conversation",type=str,help="conversation_id,example: 123") 
    start_parser.add_argument("--url",type=str,help="仅rag,loader使用,下载的URL地址")    
    start_parser.add_argument("--depth",type=int,default=1,help="仅rag使用,下载的深度，默认为1")
    start_parser.add_argument("--rag_dbname",type=str,help="保存的向量数据库，默认rag.faiss") 
    start_parser.add_argument("--chat_dbname",type=str,help="保存的对话数据库，默认chat.sqlite") 
    start_parser.add_argument("--docx",type=str,help="docx文档文件名") 
    start_parser.add_argument("--pptx",type=str,help="pptx文档文件名") 
    start_parser.add_argument("--pdf",type=str,help="pdf文档文件名") 
    start_parser.add_argument("--glob",type=str,help="当前目录下的glob匹配的文件")
    start_parser.add_argument("--websearch",type=str,choices=["tavily","duckduckgo"],help="websearch的工具")
     
    
    start_parser.add_argument("--size",type=int,help="文档分隔的size") 
     

    # start_parser.add_argument("--only_download",type=bool,default=False,help="仅下载网页html,不进行翻译。默认:False (json模式该参数不起作用)")
    # start_parser.add_argument("-s","--size",type=int,default=1500,help="切分文件的字节大小,默认:1500")
    # start_parser.add_argument("-c","--clear_error",action="store_true",help="清除task.json文件中的错误信息,默认:False")

    serve_parser = subparsers.add_parser("serve", help="启动langserve")
    serve_parser.add_argument("--host",type=str,default="0.0.0.0",help="bind host,default:0.0.0.0")
    serve_parser.add_argument("--port",type=int,default=8000,help="listen port,default::8000")
    serve_parser.add_argument("--path",type=str,default="/test",help="path,default:: /test")
    serve_parser.add_argument("--llm",type=str,help="llm,example: LLM.DEEPSEEK")    
    serve_parser.add_argument("--model",type=str,help="model,example: deepseek-chat") 
    serve_parser.add_argument("--user",type=str,help="user_id,example: alice")    
    serve_parser.add_argument("--conversation",type=str,help="conversation_id,example: 123") 
           
    args = parser.parse_args()

    #print("args====>",args)

    if args.command =="reset":
        reset(args)
        return
    
    init(args)
   
    if args.command == "start":
        start(args)
    elif args.command == "serve":
        serve(args)

if __name__ == "__main__":
   main()




from langchain_community.document_loaders import DirectoryLoader

class DirLoader():
    def __init__(self,langchainLib):
        self.langchainLib = langchainLib
        self.loader  = DirectoryLoader


from ylz_utils import Color, StringLib
from ylz_utils import FileLib
from ylz_utils import LangchainLib

from pydantic import BaseModel,Field
from typing import Literal,List
from pptx.util import Inches,Cm,Pt
from pptx.enum.shapes import MSO_SHAPE

from langchain.prompts import PromptTemplate,ChatPromptTemplate
from langchain.output_parsers import OutputFixingParser
from langchain_core.messages import HumanMessage,ToolMessage
from langchain_core.tools import tool
from langchain_core.runnables import RunnablePassthrough,RunnableLambda,RunnableParallel
from langgraph.graph import START,END,StateGraph,MessagesState
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import ToolNode
from langchain_community.document_loaders import UnstructuredFileLoader

def __agent(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message if args.message else "厦门今天天气怎么样?"
    llm = langchainLib.get_llm(llm_key,llm_model)
    prompt = langchainLib.get_prompt()
    tools = []
    tavily_tool = langchainLib.toolLib.web_search.get_tool()
    tools.append(tavily_tool)
    # langchain_docs_vectorestore = langchainLib.vectorstoreLib.faiss.load("langchain_docs.db")
    # langchain_docs_retriever = langchain_docs_vectorestore.as_retriever()
    # rag_tool = langchainLib.toolLib.rag_search.get_tool(langchain_docs_retriever,"langchain docs","langchain的文档")
    # tools.append(rag_tool)
    
    agent =  langchainLib.get_agent(llm,tools)
    res = agent.stream({"messages":[("user",message)]})
    for chunk in res:
        print(chunk , end="")
    print("\n\n",langchainLib.get_llm(full=True),"\n")

def __chat(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    message = args.message
    model = args.llm_model
    user_id = args.user if args.user else 'default'
    conversation_id = args.conversation if args.conversation else 'default'
    llm = langchainLib.get_llm(key=llm_key,model=model)
    dbname = args.chat_dbname or 'chat.sqlite'
    #### chat 模式
    prompt = langchainLib.get_prompt(use_chat=True)
    if dbname:
        langchainLib.llmLib.set_dbname(dbname)
    chat = langchainLib.get_chat(llm,prompt)
    chain = chat | langchainLib.get_outputParser()
    while True:
        if not message:
            message=input("USER:")
        else:
            print(f"USER:{message}")
        if message.lower() in ['/quit','/exit','/stop','/bye','/q']:
            print("Good bye!!")
            break
        res = chain.stream({"input":message}, {'configurable': {'user_id': user_id,'conversation_id': conversation_id}} )
        message = ""
        print("AI:",end="")
        for chunk in res:
            print(chunk,end="")
        print("\n")
    print("\n",langchainLib.get_llm(full=True))

def __llm_test(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    for _ in range(3):
        llm = langchainLib.get_llm(llm_key)
        res = llm.invoke("hello")
        print("res:",res)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])

def __outputParser_test(langchainLib:LangchainLib,args):
    message = args.message
    llm_key = args.llm_key
    class Food(BaseModel):
        name: str = Field(description="name of the food")
        place: str = Field(defualt="未指定",description="where to eat food?",examples=["家","公司","体育馆"])
        calories: float|None = Field(title="卡路里(千卡)",description="食物热量")
        amount: tuple[float,str] = Field(description="tuple of the value and unit of the food amount")
        time: str = Field(default = "未指定",description="the time of eating food,确保每个食物都有对应的时间，如果没有指定则为空字符串")
        health: Literal["健康","不健康","未知"]|None = Field(description="根据常识判断食物是否为健康食品",examples=["健康","不健康","未知"])
    class Foods(BaseModel):
        foods: List[Food]

    # Set up a parser + inject instructions into the prompt template.
    parser = langchainLib.get_outputParser(Foods)

    prompt1 = PromptTemplate(
        template="用中文分析句子的内容。如果没有指定食物热量则根据食物的名称和数量进行估计。判断食物是否为健康\n{format_instructions}\n{command}\n",
        input_variables=["command"],
        partial_variables={"format_instructions": parser.get_format_instructions()},
    )


    # And a query intended to prompt a language model to populate the data structure.
    fixed_parser = OutputFixingParser.from_llm(parser=parser, llm=langchainLib.get_llm(llm_key))
    # retry_parser = RetryOutputParser.from_llm(parser=parser, llm=langchainLib.get_llm())

    prompt2 = langchainLib.get_prompt(
            system_prompt = "用中文分析句子的内容。如果没有指定食物热量则根据食物的名称和数量进行估计。同时判断食物是否为健康",
            human_keys={"command":"输入的语句"},
            outputParser = parser,
            use_chat = False
        )

    #chain = prompt2 | langchainLib.get_llm() | fixed_parser
    
    chain = prompt2 | langchainLib.get_llm(llm_key) | parser
    
    # retry_chain = RunnableParallel(
    #     completion=chain, prompt_value=prompt1,
    #     ) | RunnableLambda(lambda x:retry_parser.parse_with_prompt(**x))

    #output = chain.invoke({"command": "我今天在家吃了3个麦当劳炸鸡和一个焦糖布丁，昨天8点在电影院吃了一盘20千卡的西红柿炒蛋"})
    output = chain.invoke({"command": message})
    #retry_parser.parse_with_prompt(output['completion'], output['prompt_value'])
    #retry_parser.parse_with_prompt(output['completion'],output['prompt_value'])
    print("\noutput parser:",output)

    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])


def __runnalble_test(langchainLib:LangchainLib):
    runnable1 = RunnableParallel(
    passed=RunnablePassthrough(),
    modified=lambda x: x["num"] + 1,
    other = RunnableParallel(
        passed=RunnablePassthrough(),
        modified=lambda x: x["num"] * 2,
        )
    )
    #runnable1.invoke({"num":1})
    def f(x):
        return x['modified']*100

    runnable2 = runnable1 | RunnableParallel(
        {"origin": RunnablePassthrough(),
        "new1": lambda x: x["other"]["passed"],
        "new2": RunnableLambda(f)}
        )

    #for trunck in runnable2.stream({"num": 1}):
    #  print(trunck)
    res = runnable2.invoke({"num": 1})
    print("\nrunnalbe:",res)

def __prompt_test(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    prompt = langchainLib.get_prompt(f"你对日历时间非常精通",human_keys={"context":"关联的上下文","question":"问题是"},use_chat = False)
    prompt.partial(context="我的名字叫小美")   # 这个为什么不起作用?
    print("!!!!",langchainLib.llmLib.default_llm_key,llm_key)
    llm = langchainLib.get_llm(llm_key,llm_model)
    print(llm)
    outputParser = langchainLib.get_outputParser()
    chain = prompt | llm | outputParser

    res = chain.invoke({"question":"我叫什么名字，今天礼拜几","context":"我是海涛"})
    print(res)
    res = chain.stream({"question":"用我的名字写一首诗歌","context":"我的名字叫小美"})
    for chunk in res:
        print(chunk,end="")
    
    prompt = PromptTemplate.from_template("用中文回答：{topic} 的{what}是多少")
    llm = langchainLib.get_llm("LLM.GROQ")
    print(llm)
    chain = prompt | llm | outputParser
    promise = chain.batch([{"topic":"中国","what":"人口"},{"topic":"美国","what":"国土面积"},{"topic":"新加坡","what":"大学"}])
    print(promise)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])

def __rag_test(langchainLib:LangchainLib,args):
    embedding_key = args.embedding_key
    embedding_model = args.embedding_model
    rag_dbname = args.rag_dbname or "rag.faiss"
    url = args.url
    pptx = args.pptx
    docx = args.docx
    pdf = args.pdf
    glob = args.glob
    message = args.message
    
    
    if (not url and not pptx and not docx and not pdf and not glob) and  (not message):
        print(f"1、指定url/pptx/docx:系统将文档下载切片后向量化到{rag_dbname}数据库\n2、指定message:系统将从{rag_dbname}数据库中搜索相关的两条记录。\n您需要至少指定url和message中的一个参数.")
        return

    if embedding_key or embedding_model:
        embedding = langchainLib.get_embedding(embedding_key,embedding_model)
    else:
        embedding = None

    if (url or pptx or docx or pdf or glob):
        docs = __loader_test(langchainLib,args)
        print("#"*60)
        if not docs:
            return
        if url and rag_dbname:
            ##### create vectorestore
            # url = "https://python.langchain.com/v0.2/docs/concepts/#tools"
            # faiss_dbname = "langchain_docs.faiss"
            print("result:",[{"doc_len":len(doc['doc'].page_content),"doc_blocks":len(doc['blocks'])} for doc in docs])
            for doc in docs:
                blocks = doc['blocks']
                vectorestore,ids = langchainLib.vectorstoreLib.faiss.create_from_docs(blocks,embedding)
                langchainLib.vectorstoreLib.faiss.save(rag_dbname,vectorestore)
                print("ids:",ids)
        else:
                vectorestore,ids = langchainLib.vectorstoreLib.faiss.create_from_docs(docs,embedding)
                langchainLib.vectorstoreLib.faiss.save(rag_dbname,vectorestore)
                print("ids:",ids)
    if message and rag_dbname:   
        # docs = [Document("I am a student"),Document("who to go to china"),Document("this is a table")]
        # vectorestore = langchainLib.vectorstoreLib.faiss.create_from_docs(docs)
        # langchainLib.vectorstoreLib.faiss.save("test.faiss",vectorestore)
        
        vectorestore = langchainLib.vectorstoreLib.faiss.load(rag_dbname,embedding)
        print("v--->",langchainLib.vectorstoreLib.faiss.search(message,vectorestore,k=2))
    
    ###### have bug when poetry add sentence_transformers   
    #v1 = langchainLib.get_huggingface_embedding()
    #print("huggingface BGE:",v1)

    ###### tet google embdeeding
    # embed = langchainLib.get_embedding("EMBEDDING.GEMINI")
    # docs = [Document("I am a student"),Document("who to go to china"),Document("this is a table")]
    # vectorestore = langchainLib.vectorstoreLib.faiss.create_from_docs(docs,embedding=embed)
    # langchainLib.vectorstoreLib.faiss.save("test.faiss",vectorestore,index_name="gemini")

def __loader_test(langchainLib:LangchainLib,args):
    url = args.url
    depth = args.depth
    docx_file = args.docx
    pptx_file = args.pptx
    pdf_file = args.pdf
    glob = args.glob
    chunk_size = args.size or 512
    only_one = list(filter(lambda x: x,[url,docx_file,pptx_file,pdf_file,glob]))
    if len(only_one) != 1:
        print(f"请指定url,docx,pptx,pdf,glob其中的一个")
        return 
    if url:
        result = langchainLib.load_url_and_split_markdown(url = url,max_depth = depth, chunk_size=chunk_size)
        print("result:",[{"doc_len":len(doc['doc'].page_content),"doc_blocks":len(doc['blocks']),"metadata":doc['metadata']} for doc in result])
    elif docx_file:
        result = langchainLib.documentLib.docx.load_and_split(docx_file,chunk_size=chunk_size)
        print(result)
    elif pptx_file:
        result = langchainLib.documentLib.pptx.load_and_split(pptx_file,chunk_size=chunk_size)
        print(result)
    elif pdf_file:
        result = langchainLib.documentLib.pdf.load_and_split(pdf_file,chunk_size=chunk_size)
        print(result)
    elif glob:
        loader_cls = UnstructuredFileLoader
        if glob.find(".docx")>=0:
            loader_cls = langchainLib.documentLib.docx.loader
        elif glob.find(".pptx")>=0:
            loader_cls = langchainLib.documentLib.pptx.loader
        elif glob.find(".pdf")>=0:
            loader_cls = langchainLib.documentLib.pdf.loader
        loader  = langchainLib.documentLib.dir.loader(".",glob=glob,show_progress=True,loader_cls=loader_cls)
        result  = loader.load_and_split(langchainLib.splitterLib.get_textsplitter(chunk_size=chunk_size,chunk_overlap=0))
        print(result)
    return result

def __tools_test(langchainLib:LangchainLib,args):
    tool = langchainLib.toolLib.python_repl.get_tool()
    res = tool.invoke(
"""
print(3+2)
""")
    print("python repl=",res)
    tool = langchainLib.toolLib.wolfram_alpha.get_tool()
    res = tool.run("what is the square root of 25?")
    print(res)
    return 
    # tool: TavilySearchResults = langchainLib.get_search_tool("TAVILY")
    # prompt = langchainLib.get_prompt(is_chat=False,human_keys={"context":"关联的上下文是:","question":"问题是:"})
    # res = tool.invoke("易联众现在股价是多少？")
    # print(res)
    # llm = langchainLib.get_llm("LLM.DEEPSEEK")
    # chain = RunnableParallel({
    #     "question": RunnablePassthrough(),
    #     "context": tool
    # }) | prompt | llm 
    # res = chain.invoke("易联众现在股价是多少？")
    # print(res)

    # print("#"*50)
    class Output(BaseModel):
        name:str = Field(description= "姓名")
        age:int = Field(description= "年龄")

    langchainLib.add_plugins()
    llm_key = args.llm_key
    llm = langchainLib.get_llm(llm_key)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])

    outputParser = langchainLib.get_outputParser(Output,fix=True,llm = llm,retry=3)
    prompt = langchainLib.get_prompt(human_keys = {"ask":"问题"},use_chat=False)
    search = langchainLib.get_search_tool("TAVILY")
    tools = [search]
    #chain  = prompt | (lambda x: x.messages[-1].content) | search | (lambda x:'```json\n{"name":中国,"age":"20"}') | outputParser
    #chain  = prompt | llm | outputParser
    chain = prompt | RunnableLambda(lambda x:x.text,name="抽取text") | search 
    #print("chain to_json = ",chain.to_json())
    #print("chain name=",chain.get_name())
    print("\nchain graph=",chain.get_graph(),"\n")
    res = chain.invoke({"ask": "北京人口多少"})
    print(type(res),res)
    #FileLib.writeFile("graph1.png",chain.get_graph().draw_mermaid_png(),mode="wb")

    # llm = langchainLib.get_llm("LLM.DEEPSEEK")
    # print("llm",llm.to_json())
    llm.bind_tools(tools)
    res = llm.invoke({"ask":"北京2024年人口多少"})
    print(res)
def __graph_test(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message
    chat_dbname = args.chat_dbname
    rag_dbname = args.rag_dbname
    user = args.user or 'default'
    conversation = args.conversation or 'default'
    thread_id = f"{user}-{conversation}"
    websearch_key = args.websearch
    if chat_dbname:                                  
        langchainLib.graphLib.set_dbname(chat_dbname)
        print("!!!",f"使用对话数据库{chat_dbname}")
    if rag_dbname:
        retriever = langchainLib.vectorstoreLib.faiss.load(rag_dbname).as_retriever()
        langchainLib.graphLib.set_ragsearch_tool(retriever)
        print("!!!",f"使用知识库{rag_dbname}")
    if websearch_key:
        langchainLib.graphLib.set_websearch_tool(websearch_key)
        print("!!!",f"使用搜索工具{websearch_key}")
        
    graph = langchainLib.get_graph(llm_key=llm_key,llm_model=llm_model)
    system_message = \
"""
请确保中文正确。只有计算和时间问题才使用python_repl工具,使用python_repl工具时要记得用print函数返回结果
不要产生幻觉，不知道的问题优先从互联网查询，关于用户自己的问题可以向用户询问。
当碰到需要需要用户来确认的问题或你需要用户告诉你的问题时，请使用使用human工具向用户询问。注意，询问时请提出询问的具体问题，不要重复我提出的问题
"""
    if not message:
        message = input("User: ")
    langchainLib.graphLib.graph_stream(graph,message,thread_id = thread_id,system_message=system_message)
    while True:
        if not message:
            message = input("User: ")
        else:
            print(f"User:{message}")
        if message.lower() in ["/quit", "/exit", "/stop","/q","/bye"]:
            print("Goodbye!")
            break
        #system_message = """请始终用中文回答"""
        langchainLib.graphLib.graph_stream(graph,message,thread_id = thread_id,system_message=system_message)
        snapshot = langchainLib.graphLib.graph_get_state(graph,thread_id)
        if snapshot.next: 
            question = snapshot.values["messages"][-1].tool_calls[0]["args"]["request"]            
            tool_call_id = snapshot.values["messages"][-1].tool_calls[0]["id"]
            message = input(f"{question}: ")
            if message.strip().lower() != '':
                tool_message = ToolMessage(tool_call_id=tool_call_id, content=message)
                langchainLib.graphLib.graph_update_state(graph,thread_id=thread_id, values = {"messages":[tool_message]})
                langchainLib.graphLib.graph_stream(graph,None,thread_id = thread_id)

        message = ""

    current_state = langchainLib.graphLib.graph_get_state(graph,thread_id)
    print("\n本次对话的所有消息:\n",current_state.values["messages"])

    langchainLib.graphLib.export_graph(graph)

def start(args):
    langchainLib = LangchainLib()
    #langchainLib.add_plugins()
    if args.mode == "llm":
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试llm {Color.RESET}")
        __llm_test(langchainLib,args)
    elif args.mode == "prompt":
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试prompt {Color.RESET}")
        __prompt_test(langchainLib,args)    
    elif args.mode == 'loader':    
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试loader {Color.RESET}")
        __loader_test(langchainLib,args)    
    elif args.mode == 'runnable':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试runnable {Color.RESET}")
        __runnalble_test(langchainLib)
    elif args.mode == 'outputParser':    
        StringLib.logging_in_box(f"{Color.YELLOW} 测试outputParser {Color.RESET}")
        __outputParser_test(langchainLib,args)
    elif args.mode == 'rag':    
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试rag {Color.RESET}")
        __rag_test(langchainLib,args)
    elif args.mode == 'tools':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试tools {Color.RESET}")
        __tools_test(langchainLib,args)
    elif args.mode == 'graph':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试graph {Color.RESET}")
        __graph_test(langchainLib,args)
    elif args.mode == 'chat':
        __chat(langchainLib,args)
    elif args.mode == 'agent':
        __agent(langchainLib,args)
    else:
        print("args=",args)
        print("llms--->:",[(item["type"],item["api_key"],item["model"],item["used"]) for item in langchainLib.get_llm(full=True)])
        print("embeddings---->:",[(item["type"],item["api_key"],item["model"],item["used"]) for item in langchainLib.get_embedding(full=True)])
        #loader = langchainLib.documentLib.pptx.loader("30335320.pptx")
        #docs = loader.load()
        #print(docs)
        return
        loader = langchainLib.documentLib.pptx.newer("test.pptx")
        loader.add_slide(0).set_title("Hello World","gogogo").add_text("youht",10,10,60,40)
        tab = [{"name":"youht","age":20},{"name":"jinli","age":10}] 
        tx = loader.add_slide(1).set_title("你好","step1").add_text("Step1",10,10,100,40)
        loader.add_text_paragraph(tx,"如何学习python",size=30,level=1)
        loader.add_text_paragraph(tx,"numpy",bold=True,size=20,level=2)
        loader.add_slide(2).set_title("你好","step1").add_table(tab,100,100,600,400,with_header = True)
        shape = loader.add_slide(8).set_title("Shape").add_shape(MSO_SHAPE.ROUNDED_RECTANGLE,50,50,50,50,
                                                        line_brightness=0.5,
                                                        text="hello")
        print(loader.get_shape_sizes(shape))
        loader.add_slide().set_title("chart1").add_chart("pie",
                                                        [{"s1":{"a":1,"b":2,"c":3}},
                                                         {"s2":{"a":3,"b":2,"c":1}}],10,10,600,300)
        loader.add_slide().set_title("chart2").add_chart("bubble",
                                                        [{"s1":[(1,2,10),(4,6,3),(2,3,1)]},
                                                         {"s2":[(2,1,2),(8,2,3),(4,1,5)]}],10,10,600,300)

        loader.save()


from io import BytesIO
import imghdr
from PIL import Image
import base64
import requests
import logging

import xml.etree.ElementTree as ET

def get_url_image(url):
    # 从 URL 下载图片
    response = requests.get(url)
    # 确保请求成功
    if response.status_code == 200:
        # 检查响应头的 Content-Type
        content_type = response.headers.get('Content-Type', '')
        if content_type.startswith('image/'):
            image_type = imghdr.what(None, h=response.content)
            if image_type in ['jpeg', 'jpg', 'bmp', 'png','webp']:
                # 将下载的图片数据转换为字节流
                image_data = BytesIO(response.content)
                # 使用 Pillow 的 Image 模块打开图片
                image = Image.open(image_data)
                # 显示图片
                #display(image)
                return (image,image_type)
            else:
                if content_type.index("svg") > -1:
                    return (response.content, "svg")
                else:
                    logging.debug(f"无法识别图片类型：{image_type}")
                    return (None, None)
    else:
        logging.debug(f"无法下载图片，状态码：{response.status_code}") 
        return (None, None)

def convert_to_base64(pil_image):
    buffered = BytesIO()
    pil_image.save(buffered, format="PNG")  # You can change the format if needed
    img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")
    return img_str

def indent_svg(elem, level=0):
    """
    Helper function to indent the XML for pretty printing.
    """
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for subelem in elem:
            indent_svg(subelem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


from langgraph.prebuilt import chat_agent_executor

class AgentLib():
    def __init__(self,langchainLib):
        self.langchainLib = langchainLib
    def get_agent(self,llm,tools):
        return chat_agent_executor.create_function_calling_executor(llm,tools)



from ylz_utils.langchain.tools.python_repl import PythonREPLTool
from ylz_utils.langchain.tools.rag_search import RagSearchTool
from ylz_utils.langchain.tools.web_search import WebSearchTool
from ylz_utils.langchain.tools.wolfram_alpha import WolframAlphaTool


class ToolLib():
    def __init__(self,langchainLib):
        self.langchainLib = langchainLib
        self.web_search = WebSearchTool(langchainLib)
        self.rag_search = RagSearchTool(langchainLib)
        self.python_repl = PythonREPLTool(langchainLib)
        self.wolfram_alpha = WolframAlphaTool(langchainLib)


import random
import time

from ylz_utils import Config

from langchain_together import TogetherEmbeddings
from langchain_google_genai import GoogleGenerativeAIEmbeddings
from langchain_ollama import OllamaEmbeddings

from langchain_community.embeddings import (
    HuggingFaceEmbeddings,
    HuggingFaceBgeEmbeddings,
)

class EmbeddingLib():
    embeddings:list  = []
    default_embedding_key = None
    def __init__(self):
        self.config = Config.get()
        self.regist_embedding()
        if self.config.get("EMBEDDING.DEFAULT"):
            default_embedding_key = self.config.get("EMBEDDING.DEFAULT")
            self.set_default_embedding_key(default_embedding_key)
        else:
            self.clear_default_embedding_key()
    def set_default_embedding_key(self,key):
        embeddings = [item for item in self.embeddings if item['type']==key]
        if embeddings:
            self.default_embedding_key = key
        else:
            self.clear_default_embedding_key()
    def get_default_embedding_key(self):
        return self.default_embedding_key
    def clear_default_embedding_key(self):
        self.default_embedding_key=None

    def get_embedding(self,key=None,model=None, full=False) :
        if full:
            return self.embeddings
        if self.embeddings:
            if not key:
                if self.default_embedding_key:
                    embeddings = [item for item in self.embeddings if item['type']==self.default_embedding_key]
                else:
                    embeddings = self.embeddings
            else:
                embeddings = [item for item in self.embeddings if item['type']==key]
            if embeddings:
                embedding = random.choice(embeddings)    
                if not embedding.get('embedding'):
                    embed_type = embedding['type']
                    if  embed_type == 'EMBEDDING.TOGETHER':
                        embedding['embedding'] = TogetherEmbeddings(api_key = embedding.get('api_key'),model=embedding.get('model'))
                    elif embed_type == "EMBEDDING.GEMINI" :
                        embedding['embedding'] = GoogleGenerativeAIEmbeddings(model=embedding.get('model'),google_api_key=embedding.get('api_key'))
                    elif embed_type == 'EMBEDDING.OLLAMA':
                        embedding['embedding'] = OllamaEmbeddings(model=embedding.get('model'))
                    else:
                        raise Exception(f"目前不支持{embedding['type']}嵌入模型")
                embedding['used'] = embedding.get('used',0) + 1 
                embedding['last_used'] = time.time()
                return embedding['embedding']
        if self.default_embedding_key:
            raise Exception(f"请确保{self.default_embedding_key}_API_KEYS环境变量被正确设置,然后调用regist_embedding注册语言模型")
        else:
            raise Exception(f"请确保<EMBEDDING PROVIDER>_API_KEYS环境变量被正确设置,然后调用regist_embedding注册语言模型")
    
    def regist_embedding(self):
        defaults = {
                      "EMBEDDING.TOGETHER": {"model":"BAAI/bge-large-en-v1.5"},
                      "EMBEDDING.GEMINI": {"model":"models/embedding-001"},
                      "EMBEDDING.OLLAMA": {"model":"mxbai-embed-large"}
                  }
        for key in defaults:
            default = defaults[key]
            embed = self.config.get(key)
            if not embed:
                continue
            base_url = embed.get("BASE_URL")
            api_keys = embed.get("API_KEYS")
            if api_keys:
                api_keys = api_keys.split(",")
                if not api_keys[-1]:
                    api_keys.pop()
            else:
                api_keys = []
            model= embed.get("MODEL") if embed.get("MODEL") else default['model']
            for api_key in api_keys:
                self.embeddings.append({
                    "embedding": None,
                    "type": key,
                    "base_url": base_url,
                    "api_key":api_key,
                    "model":model,
                    "used":0,
                    "last_used": 0 
                })
                
    def get_huggingface_embedding(self,mode="BGE"):
        if mode=="BGE":
            embedding = HuggingFaceBgeEmbeddings()
        else:
            embedding = HuggingFaceBgeEmbeddings()
        return embedding

